# Copyright (c) 2009, Evan Fosmark
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.
'''Cross platform mechanism to acquire and release a file lock. The module
uses additional lock file to mark if the file is currently used.

Example:
from filelock import FileLock
with FileLock('path_to_my_file'):
    # My operations with this files comes here
'''
import os
import logging

logger = logging.getLogger(__name__)

class SecureOpen(object):
    """ Implementation of cross platform secure file writer logic. Use this
    class for write file operations only. The class is responsible to securely
    persist the file content into a file. In case we are running on Windows we
    may fail because the file could be opened by another process. If this is
    the case, then we have to:
        1) move preliminary known file path to temporary location
        2) delete preliminary known file path
        3) Write the file content to preliminary known file path.
    """
    def __init__(self, filePath, openFileFunction=open, strict=False,
                 *openFileArgs, **openFileKwArgs):
        """ Creates secure file handle.

            @param filePath: File wanted to be opened
            @type filePath: str

            @param openFileFunction: Function which will be used to open the file.
              Default is os.open
            @param openFileFunction: function

            @param strict: Specifies the behavior if file content cannot be
              persisted.
                  False(default) - Warn and ignore the exceptions
                  True - Throw exception because file content cannot be persisted
        """
        self.filePath = filePath
        self.newFilePath = "%s.new" % (filePath)
        self.oldFilePath = "%s.old" % (filePath)
        self.newFileDescriptior = None
        self.strict = strict
        self.openFileFunnction = openFileFunction
        self.openFileArgs = openFileArgs
        self.openFileKwArgs = openFileKwArgs

    def open(self):
        """Opens the file. Actually it will open self.newFile and will write
        the content there. We do that to decrease the chance two processes to
        use the same file at the same time.
        """
        if self.newFileDescriptior is not None:
            error = 'File %s is already opened' % self.newFilePath
            logger.error(error)
            raise ValueError(error)

        self.newFileDescriptior = self.openFileFunnction(self.newFilePath,
                                                         *self.openFileArgs,
                                                         **self.openFileKwArgs)
        return self.newFileDescriptior

    @staticmethod
    def moveFile(srcFile, dstFile):
        '''Moves the source file to destination file. If destination file
        exists then the file will be overwritten.

        @param srcFile: Path to the source file which has to be renamed
        @type srcFile: str

        @param dstFile: Destination file path
        @type dstFile: str

        @raise IOError: If move operation failed to be performed.
        '''
        os.rename(srcFile, dstFile)

    @staticmethod
    def renameFile(srcFile, dstFile, helperFile):
        '''Tries to rename the source file to destination file. If the rename
        operation failed(perhaps because the destination file is currently in
        use), then try to:
        1/ Rename destination file to helper file in order to release the
          destination file
        2/ Retry to rename the source file to destination file.

        @param srcFile: Path to the source file which has to be renamed
        @type srcFile: str

        @param dstFile: Destination file path
        @type dstFile: str

        @param helperFile: File used for intermediate rename transaction
        @type helperFile: str

        @raise IOError: If rename operation failed to be performed.
        '''
        try:
            SecureOpen.moveFile(srcFile, dstFile)
        except IOError:
            try:
                SecureOpen.moveFile(dstFile, helperFile)
            except IOError, e:
                # This is fine, perhaps the destination file is already gone
                logger.info("%s", e)
            SecureOpen.moveFile(srcFile, dstFile)

    def close(self):
        """ Releases the self.newFileDescriptor and move the file content to
        preliminary set file location.
        """
        if self.newFileDescriptior is None:
            return

        self.newFileDescriptior.close()
        self.newFileDescriptior = None

        # Rename the temporary file path to the wanted destination path.
        try:
            SecureOpen.renameFile(self.newFilePath, self.filePath,
                                  self.oldFilePath)
        except Exception, e:  # pylint: disable=W0703
            logger.exception('Failed to create a file %s. Cause: %s',
                             self.filePath, str(e))
            if self.strict:
                raise

    def __getattr__(self, name):
        # Forward the rest of the calls to newFileDescriptor
        return getattr(self.newFileDescriptior, name)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, _type, _value, _traceback):
        self.close()

    def __del__(self):
        self.close()
