# Copyright 2015 VMware, Inc.  All rights reserved. -- VMware Confidential
#
''' Module containing vCSA specific utilities.
'''
import os
from os.path import join, normpath
import json
import logging
from msgL10n import MessageMetadata
from l10n import msgMetadata as _T, localizedString as _
from patch_errors import PermanentError
from patch_specs import DiscoveryResult
import featureState

SERVICE_NAME_L10N_ID = "l10n.componentName"
DEPLOYMENT_TYPE_FILE = "/etc/vmware/deployment.node.type"

EMBEDDED_DEPLOYMENT_TYPE = "embedded"
MANAGEMENT_DEPLOYMENT_TYPE = "management"
INFRASTRUCTURE_DEPLOYMENT_TYPE = "infrastructure"

__serviceConfig = None
logger = logging.getLogger(__name__)

def _getSdkConfig(configFileName):
    '''Gets SDK configuration file

    @param configFileName: Name of the configuration file the caller is
      interested in
    @type configFileName: str

    @return: Absolute path to the file
    @rtype: str
    '''
    # SDK module configurations reside in $(SDK)/config
    return normpath(join(__file__, os.pardir, 'config', configFileName))

def getDeploymentType(deploymentTypeFile=DEPLOYMENT_TYPE_FILE):
    '''Extracts the vCSA deployment type. The deplopyment type result could be
    either embedded, management or infrastructure.

    @param deploymentTypeFile: File where vCSA deployment type information is
      persisted on the disk
    @type deploymentTypeFile: str

    @raise PermenentError: if the deployment type cannot be retrieved.
    '''
    if os.path.exists(deploymentTypeFile):
        with open(deploymentTypeFile) as fp:
            return fp.read().strip()
    else:
        raise PermanentError(cause=_(_T("patch.error.deploymentType.text",
                                        "vCSA Deployment type cannot be retrieved")),
                             resolution=_(_T("patch.error.deploymentType.resolution",
                                             "vCSA patching is not executed on the "
                                             "right host or the vCSA is not properly "
                                             "configured")))

def _loadServiceConfig(serviceConfigFile=_getSdkConfig("services.json")):
    '''Loads the service configuration from given configuration file. The content
    of the file is expected to be json.

    @param serviceConfigFile: Location of service configuration file
    @type serviceConfigFile: str

    @return The service config content as dictionary
    @rtype: dict

    @raise ValueError: If the file does not exist.
    '''
    if not os.path.exists(serviceConfigFile):
        raise ValueError("Invalid configuration. Cannot find services "
                         "configurations %s",serviceConfigFile)

    with open(serviceConfigFile) as fp:
        serviceData = json.load(fp)
    return serviceData

def loadComponentServiceData(serviceId):
    '''Loads the components service data for given serviceId from services.json
    file coming with the latest rpm.

    @param serviceId: The service identificator for which the data has to be
      loaded for
    @type serviceId: str

    @return: A dictionary expressing the service data for given service id
      Example: {
        "enabled": true,
        "dependsOn": ["vmware-cis-config", "visl-integration"],
        "windows": {
            "serviceName": "VMWareAfdService",
            "fbScripts": ["vmafd-firstboot.py"],
            "installFiles": ["vmware-afd.msi"]
        },
        "appliance": {
            "serviceName": "vmafdd",
            "fbScripts": ["vmware-vmafd/firstboot/vmafd-firstboot.py"],
             "installFiles": ["vmware-afd-@version@.rpm"]
        },
        "serviceDisplayName": "VMware Authentication Framework",
        "deploymentType": ["all"],
        "isCorePnidService": true
    }
    @rtype: dict

    @raise ValueError: If the serviceId does not exist
    '''
    global __serviceConfig
    if __serviceConfig is None:
        __serviceConfig = _loadServiceConfig()


    if not __serviceConfig.get(serviceId):
        raise ValueError("Invalid serviceId: %s", serviceId)

    return __serviceConfig[serviceId]

def _getSupportedDeploymentTypes(serviceData):
    '''Extract and normalize the supported deployment type for a service in the
    service data

    @param serviceData: The data for a service in the services.json file
    @type serviceData: dictionary

    @return: List of supported deployment types for service
    '''
    if serviceData["deploymentType"] == ["all"]:
        # Expand deployment types if "all" is specified.
        return [EMBEDDED_DEPLOYMENT_TYPE, MANAGEMENT_DEPLOYMENT_TYPE,
                          INFRASTRUCTURE_DEPLOYMENT_TYPE]
    else:
        return serviceData["deploymentType"]

isFeatureSwitchInitialized = False
def isFswitchEnabled(serviceId):
    """
    Some services are behind a feature switch and may not be available
    when the feature switch is off. Hardcode these here till we figure
    out a better way to get this info.
    """
    global isFeatureSwitchInitialized
    if not isFeatureSwitchInitialized:
        featureState.init(enableLogging=False)
        isFeatureSwitchInitialized = True

    #For now only VCHA switch is causing problems
    if serviceId in ['vcha', 'vmonapi', 'vmware-vmon']:
        return featureState.getVCHA()
    elif serviceId in ['netdumper-web']:
        return not featureState.getVCHA()
    elif serviceId in ['statsmonitor']:
        return featureState.getApplianceMonitoring()
    else:
        return True

def _isEnabledService(serviceId, componentData, localDeploymentType):
    '''
    Checks if the service provided is enabled on the machine. At the moment it
    checks deployment type, feature flags, and if service is enabled
    '''
    return localDeploymentType in _getSupportedDeploymentTypes(componentData) and \
            isFswitchEnabled(serviceId) and \
            componentData['enabled']


def _getDependentServices(serviceId, localDeploymentType, cachedServices = None):
    '''Returns all component dependent services. The function iterates over all
    dependent components recursively and provide the dependent component services.
    The recursion goes over Depth-First-Search in order to provide proper service
    dependency chain.

    @param serviceId: The service identificator for which the data has to be
      loaded for
    @type serviceId: str

    @param localDeploymentType: The deployment type of the vCSA
    @type localDeploymentType: str

    @return: List of dependent services in the right dependent order, i.e.
    a[i] could depend on a[j] if and only if i > j
    @rtype: list
    '''
    if cachedServices is None:
        cachedServices = {}

    class _OrderedSet(list):
        '''Due to the lack of OrderSet in python, implement dummy one by supporting
        only part of Set interface.
        Note: Do not use OrderDict because it does not exist in python2.6
        '''
        def add(self, elem):
            if elem not in self:
                self.append(elem)

        def update(self, *args, **kwargs):
            if kwargs:
                raise TypeError("update() takes no keyword arguments")

            for col in args:
                for elem in col:
                    self.add(elem)

    serviceDependencies = _OrderedSet()
    serviceData = loadComponentServiceData(serviceId)
    componentDependencies = serviceData["dependsOn"]
    for componentDependencyId in componentDependencies:
        componentDependData = loadComponentServiceData(componentDependencyId)
        if componentDependencyId in cachedServices:
            serviceDependencies.update(cachedServices[componentDependencyId])
        else:
            #Haven't see that service before compute it dependent services and add them
            # Add recursively the dependent services on our dependent component
            svces = _getDependentServices(componentDependencyId, localDeploymentType, cachedServices=cachedServices)
            serviceDependencies.update(svces)
            cachedServices[componentDependencyId] = svces

        if _isEnabledService(componentDependencyId, componentDependData, localDeploymentType):
            # Add the dependent component service
            if componentDependData["appliance"].get("serviceName"):
                serviceDependencies.add(componentDependData["appliance"]["serviceName"])

    return list(serviceDependencies)

def getComponentDefintion(serviceId):
    '''Retrieves component definition based on its service definition in
    component services json file.

    @param serviceId: The service identificator for which the data has to be
      loaded for
    @type serviceId: str

    @return: The component definition
    @rtype: _ComponentDefinition

    @raise ValueError: If the serviceId does not exist
    '''
    class _ComponentDefinition():
        def __init__(self, displayName, serviceName, serviceDependencies,
                     deploymentType):
            '''Represents component definition read by services.json file

            @param displayName: A component display name
            @type displayName: str

            @param serviceName: A component service
            @type serviceName: str

            @param serviceDependencies: List of component services dependencies
            @type serviceDependencies: list

            @param deploymentType: Deployments in which the component will be
              installed
            @type deploymentType: list
            '''
            self.displayName = displayName
            self.serviceName = serviceName
            self.serviceDependencies = serviceDependencies
            self.deploymentType = deploymentType

    serviceData = loadComponentServiceData(serviceId)

    displayName = serviceData.get("serviceDisplayName")
    serviceName = serviceData["appliance"].get("serviceName")
    deploymentType = _getSupportedDeploymentTypes(serviceData)

    serviceDependencies = _getDependentServices(serviceId, getDeploymentType())

    return _ComponentDefinition(displayName, serviceName, serviceDependencies,
                                deploymentType)

def getComponentDiscoveryResult(serviceId, **kwargs):
    '''Retrieves component discovery result based on the definition in component
    services json file

    @param serviceId: The service identificator for which the discovery result
      has to be returned for
    @type serviceId: str

    @param kwargs: Arguments which will be added on top of component definition
      in order to form the Discovery Result. For conflicted arguments, the
      user given arguments will be preferred.

    @return: DiscoveryResult for given service id or None if the component should
      not be patched
    @rtype: patch_specs.DiscoveryResult

    @raise ValueError: If the serviceId does not exist
    '''
    compDef = getComponentDefintion(serviceId)
    localDeploymentType = getDeploymentType()
    patchable = localDeploymentType in compDef.deploymentType
    if not patchable:
        logger.info("Component %s is not part of deployment type %s, hence its "
                    "patch logic will not be applied", serviceId, localDeploymentType)

    displName = _(MessageMetadata(SERVICE_NAME_L10N_ID, compDef.displayName)) \
                        if compDef.displayName else None
    patchServices = [compDef.serviceName] if compDef.serviceName else []
    discoveryArgs = {
        "displayName" : displName,
        "patchServices" : patchServices,
        "dependentServices" : compDef.serviceDependencies,
        "patchable" : patchable,
    }
    discoveryArgs.update(kwargs)
    return DiscoveryResult(**discoveryArgs)