import os
import logging
from multiprocessing import Process, Queue
from Queue import Empty

from vmware.patching.utils.logging_utils import _LOGGING_FMT
from json_utils import JsonSerializer
from patch_errors import BaseError, InternalError

from vmware.patching.executor import hook_executor

LOG_LEVEL = "%(levelname)s"

def _setupLoggingFor(patchModulePath, hook):
    '''Prefix the patching hook logging with the component signature.

    @param patchModulePath: Python module where component patching hooks are defined at
    @type patchModulePath: str

    @param hook: The patching hook which has to be executed. Allowed values are
      defined at extensions.Hook structure.
    @type hook: str
    '''
    if LOG_LEVEL not in _LOGGING_FMT:
        raise ValueError("Unnable to configure the component hook logging")

    signature = "%s:%s" % (os.path.basename(patchModulePath), hook)
    # %(asctime)s.%(msecs)dZ MOD_NAME:HOOK_NAME %(levelname)s %(name)s %(message)s
    log_format = _LOGGING_FMT.replace(LOG_LEVEL, signature + LOG_LEVEL)

    for hdlr in logging.root.handlers:
        fmt = logging.Formatter(log_format)
        hdlr.setFormatter(fmt)

def startPatchProcess(queue, patchModulePath, hook, args):
    '''The newer process entry point. The method is responsible to execute the
    patching hook and return push its result in to the queue.

    @param queue: A queue where the hook result should be pushed to. The hook
      result is consisted of tuple of (HookResult and Exception) in json serialized
      string.
    @type queue: multiprocessing.Queue

    @param patchModulePath: Python module where component patching hooks are defined at
    @type patchModulePath: str

    @param hook: The patching hook which has to be executed. Allowed values are
      defined at extensions.Hook structure.
    @type hook: str

    @param args: The patching hooks input. The object will be passed to the
      the component patching hook.
    @type args: Serializable
    '''
    _setupLoggingFor(patchModulePath, hook)

    result = None,
    ex = None
    try:
        result = hook_executor.executeHook(patchModulePath, hook, args)
    except BaseError, e:
        ex = e

    json_result = JsonSerializer().serialize([result, ex])
    queue.put(json_result)

def executeHook(patchModulePath, hook, args):
    '''Executes a component patch hook in a separate python process. The routine
    waits until the process complete.

    @param patchModulePath: Python module where component patching hooks are defined at
    @type patchModulePath: str

    @param hook: The patching hook which has to be executed. Allowed values are
      defined at extensions.Hook structure.
    @type hook: str

    @param args: The patching hooks input. The object will be passed to the
      the component patching hook.
    @type args: Serializable

    @return: The result of the patching hook
    @rtype: Serializable

    @raise: patch_errors.BaseError: If the component fails to be executed in
      the provided contract.
    '''
    queue = Queue()
    p = Process(target=startPatchProcess, args=(queue, patchModulePath, hook, args))
    p.start()
    # Blocks until the process does not complete
    # TODO iradev: Kill the process if does not complete in some time
    p.join()

    try:
        json_result = queue.get_nowait()
        result, ex = JsonSerializer().deserialize(json_result)

        if ex:
            raise ex

        return result
    except Empty:
        # Thrown if the CPO hook_execution logic fail
        raise InternalError("Cannot execute the component hook")
