# Copyright 2015 VMware, Inc.  All rights reserved. -- VMware Confidential
#
'''
Script responsible to setup environment and execute the patching hook of vCSA
components. The script calls only those components which have been discovered
in prior discovery phase. The patch phase respect the component dependencies
defined in @Discovery hook, and execute the component patch hook in the right
order as pay attention to start the dependent services, before the hook is
executed.
'''
import os
import logging

from extensions import Hook
from os_utils import getCommandExitCode
from service_manager import getServiceManager

from vmware.patching.config import settings
from vmware.patching.data.model import PatchPhaseContext
from vmware.patching.executor.execution_facade import executeComponentHook
from vmware.patching.utils.reporting_utils import setupReporting

logger = logging.getLogger(__name__)

def _startDependentServices(c):
    '''Start a component dependent services. Assume those services are either
    new or already patched.

    @param c: A component, which dependent services has to be started
    @type c: vmware.patching.data.model.Component
    '''
    depServices = c.discoveryResult.dependentServices
    logger.info("Start component %s dependent services - %s", c.name,
                ", ".join(depServices))

    serviceManager = getServiceManager()
    for depService in depServices:
        serviceManager.start(depService)

def _patchComponents(ctx, userData):
    '''Execute the patching workflow. The patching workflow first start all
    component dependent services and then call the component Patch hook.

    @param ctx: Global CPO patch phase context.
    @type ctx: vmware.patching.data.model.PatchPhaseContext

    @param userData: Customer input, as result of component questions raised
      in @Discovery patching hook.
    @type userData: dict

    @return: True if patch process complete properly, and False otherwise
    @rtype bool

    @raise Exception: if patch hook execution fails.
    '''
    for c in ctx.getPatchableComponents(sort=True):
        logger.info("Patch component %s", c.name)

        # Start all dependent services before the patch hook is being called out
        _startDependentServices(c)

        executeComponentHook(Hook.Patch, ctx, c, userData)

    return True

def _startAllVMwareServices():
    '''Starts all enabled VMware services.

    @return: True if the operation succeed and False otherwise
    @rtype: bool
    '''
    command = ['/bin/service-control',
                   '--start',
                   '--all']

    logDir = settings.loggingData.directory
    outFileName = os.path.join(logDir, 'startAllServices.out.log')
    errFileName = os.path.join(logDir, 'startAllServices.err.log')

    logger.info("Starting all VMware services... The immediate command stdout "
                "is redirected to file %s and stderr is redirected to file %s",
                outFileName, errFileName)
    exitCode = getCommandExitCode(command,
                                  localStdoutFile=outFileName,
                                  localStderrFile=errFileName)
    with open(outFileName) as fp:
        out = fp.read()

    with open(errFileName) as fp:
        err = fp.read()

    logger.info("Start All VMware services: command=%s, exit-code=%s, "
                "stdout=%s, stderr=%s", command, exitCode, out, err)

    return exitCode == 0

def _updateVMwareServicesDependencies():
    '''Updates VMware service dependencies.

    @return: True if the operation succeed and False otherwise
    @rtype: bool
    '''
    command = ['/bin/update-svc-deps']
    exitCode = getCommandExitCode(command)

    logger.info("Update service dependencies command completed with exit code %s",
                exitCode)
    return exitCode == 0

def patch(stageDir, userData, outputFile):
    '''The entry point of patch CPO phase.

    The patch phase is responsible to setup environment and execute the patching
    hook of vCSA components. The patch phase calls only those components which
    have been discovered in prior discovery phase. The patch phase respect the
    component dependencies defined in @Discovery hook, and execute the component
    patch hook in the right order as pay attention to start the dependent
    services, before the hook is executed.

    @param stageDir: global CPO stage directory. All components stage
      directories will be created as sub-directories of global stage directory.
    @type stageDir: str

    @param userData: Customer input, as result of component questions raised
      in @Discovery patching hook.
    @type userData: dict

    @param outputFile: File where the output of the command needs to be written
      to
    @type outputFile: str

    @return: True if the patch phase succeed and False otherwise
    @rtype: bool

    @raise Exception: if patch phase raise unhandled exception
    '''
    try:
        setupReporting(outputFile)

        ctx = PatchPhaseContext.load(stageDir)
        _patchComponents(ctx, userData)

        # TODO iradev: Move the code below to a common patching hook when
        # all components plugin the framework. Otherwise the dependencies
        # cannot be specified up front.
        result = _startAllVMwareServices() and _updateVMwareServicesDependencies()
        return result
    except Exception:
        # Persist error in the file if needed
        raise