# Copyright 2015 VMware, Inc.  All rights reserved. -- VMware Confidential
#
'''
Script responsible to setup environment and execute the pre-patching hook of vCSA
components. The script calls only those components which have been discovered
in prior discovery phase.
'''
import os
import logging

from extensions import Hook
from os_utils import getCommandExitCode

from vmware.patching.config import settings
from vmware.patching.data.model import PatchPhaseContext
from vmware.patching.executor.execution_facade import executeComponentHook
from vmware.patching.utils.reporting_utils import setupReporting

logger = logging.getLogger(__name__)

def _stopAllVMwareServices():
    '''Stops all VMware services defined in legacy vCSA version.

    @return: True if the operation succeed and False otherwise
    @rtype: bool
    '''
    command = ['/bin/service-control',
                   '--stop',
                   '--all']

    logDir = settings.loggingData.directory
    outFileName = os.path.join(logDir, 'stopAllServices.out.log')
    errFileName = os.path.join(logDir, 'stoptAllServices.err.log')

    logger.info("Stopping all VMware services... The immediate command stdout "
                "is redirected to file %s and stderr is redirected to file %s",
                outFileName, errFileName)
    exitCode = getCommandExitCode(command,
                                  localStdoutFile=outFileName,
                                  localStderrFile=errFileName)
    with open(outFileName) as fp:
        out = fp.read()

    with open(errFileName) as fp:
        err = fp.read()

    logger.info("Stop All VMware services: command=%s, exit-code=%s, "
                "stdout=%s, stderr=%s", command, exitCode, out, err)

    return exitCode == 0

def _prepatchComponents(ctx, userData):
    '''Executes the components Prepatch hook.

    @param ctx: Global CPO patch phase context.
    @type ctx: vmware.patching.data.model.PatchPhaseContext

    @param userData: Customer input, as result of component questions raised
      in @Discovery patching hook.
    @type userData: dict

    @return: True if prepatch process complete properly, and False otherwise
    @rtype bool

    @raise Exception: if prepatch hook execution fails.
    '''
    for c in ctx.getPatchableComponents(sort=False):
        logger.info("Prepatch component %s", c.name)

        executeComponentHook(Hook.Prepatch, ctx, c, userData)

    return True

def prepatch(stageDir, userData, outputFile):
    '''The entry point of prepatch CPO phase.

    The prepatch phase is responsible to setup environment and execute the
    prepatching hook of vCSA components. The patch phase calls only those
    components which have been discovered in prior discovery phase. The prepatch
    phase stops all VMware services before vCSA prepatch hook is executed.

    @param stageDir: global CPO stage directory. All components stage
      directories will be created as sub-directories of global stage directory.
    @type stageDir: str

    @param userData: Customer input, as result of component questions raised
      in @Discovery patching hook.
    @type userData: dict

    @param outputFile: File where the output of the command needs to be written
      to
    @type outputFile: str

    @return: True if the prepatch phase succeed and False otherwise
    @rtype: bool

    @raise Exception: if prepatch phase raise unhandled exception
    '''
    try:
        setupReporting(outputFile)

        ctx = PatchPhaseContext.load(stageDir)

        # TODO iradev: Move the code below to a common patching hook when
        # all components hook in the framework. Otherwise the dependencies
        # cannot be specified
        result = _stopAllVMwareServices()
        if result:
            _prepatchComponents(ctx, userData)

        return result
    except Exception:
        # Persist error in the file if needed
        raise
