#!/usr/bin/env python
# Copyright 2014 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""
    This script is executed by the `software-packages stage` command.
    Should this script exit with 1, the stage operation will fail.

    The purpose of this script is to modify the content that gets
    staged on the appliance. For most appliances, all the patch
    payload rpms are applicable, so the pre-stage script should be
    very simple.

    All the patching installation related logic should be placed in
    the pre-install.sh and post-install.sh scripts.
"""
import copy
import json
import os
import sys
import shutil
import subprocess

join = os.path.join
DEPLOYMENT_TYPE_FILE = "/etc/vmware/deployment.node.type"
IGNORELISTPATTERN = "%s_ignore.lst"

# If you want to skip a certain rpm during a patch or
# update, you can populate this variable with those
# rpms. Ideally this should be a configuration file.
PATCHIGNORELIST = [""]


def generateRPMBlacklist():
    """
        - Copy the correct ignore list file as rpm-blacklist.lst
        - Add rpms from PATCHIGNORELIST to rpm-blacklist.lst
    """
    try:
        if len(sys.argv) < 3:
            print "Invalid number of args for prestage script %r" % sys.argv
            return 1

        deployType = ""
        if os.path.exists(DEPLOYMENT_TYPE_FILE):
            with open(DEPLOYMENT_TYPE_FILE, 'r') as f:
                deployType = f.read().strip()

        print "CloudVM Type: %s" % deployType
        blacklistFile = join(sys.argv[2], "rpm-blacklist.lst")
        if deployType:
            ignoreFile = join(sys.argv[1], IGNORELISTPATTERN % (deployType))
            shutil.copyfile(ignoreFile, blacklistFile)

        print "Ignore List: %r" % PATCHIGNORELIST
        if PATCHIGNORELIST:
            with open(blacklistFile, 'a') as f:
                for rpm in PATCHIGNORELIST:
                    f.write("%s\n" % rpm)

        return 0
    except Exception as  e:
        print "Failed generate rpm blacklist {%s}" % e
        return 1

def prepareManifest():
    """
        Uses an appropriate RPM ignore list to generate a MANIFESTFILE
        inside STAGEDIR.
    """
    MOUNTPOINT = "/mnt/iso-contents"
    STAGEDIR = "/var/software-packages"
    MANIFESTFILE = "rpm-manifest.json"
    IGNORELISTSDIR = join(MOUNTPOINT, 'extras', 'ignorelists')
    deployType = ""

    try:
        if os.path.exists(DEPLOYMENT_TYPE_FILE):
            with open(DEPLOYMENT_TYPE_FILE, 'r') as f:
                deployType = f.read().strip()
    except IOError, e:
        return 1

    try:
        ignoreList = PATCHIGNORELIST
        if deployType:
            ignoreFile = join(IGNORELISTSDIR, IGNORELISTPATTERN % (deployType))
            for line in open(ignoreFile, 'r').readlines():
                ignoreList.append(line.strip())

        rpmManifest = {}
        with open(join(MOUNTPOINT, MANIFESTFILE), 'r') as fp:
            rpmManifest = json.load(fp)

        filteredManifest = copy.deepcopy(rpmManifest)
        # Manifest file must have the expected format
        for fname, finfo in rpmManifest['files'].iteritems():
            if (finfo['relativepath'] in ignoreList) :
                del filteredManifest['files'][fname]

        destManifestPath = join(STAGEDIR, MANIFESTFILE)
        with open(destManifestPath, 'w') as destFp:
            json.dump(filteredManifest, destFp)
    except Exception, e:
        return 1
    return 0

if __name__ == '__main__':
    """ The pre stage script is specific to VCSA appliance,
        because VCSA has different deployment types/modes
        for a single appliance deliverable. The purpose of
        this pre-stage script is to filter out rpms based
        on the deployment node type on which this script is
        being run. This script should be only extended to
        add specific rpms that need to be skipped in the
        PATCHIGNORELIST.

        In Vsphere 6.0 GA release, ISO based patching was the
        only supported means to do patching. This script was
        written specifically for ISO based patching.

        Since URL patching came into picture later, the patching
        framework had to be changed to support both. This
        resulted in old pre-stage script logic not compatible
        with both ISO and URL patching.

        prepareManifest() is pre-stage script for 6.0 GA which
        work with patching framework in GA, which is ISO.

        generateRPMBlacklist() is pre-stage script for 6.0 U1
        release and forward which works with both ISO and URL.

        In 6.0 GA, framework launches this script as, pre-stage.py
        In 6.0 U1, it launches as, 'pre-stage.py srcdir destdir'
    """
    if len(sys.argv) > 1:
        # 6.0 U1+ patching framework
        status = generateRPMBlacklist()
    else:
        # 6.0 GA only patching framework
        status = prepareManifest()
    sys.exit(status)
