"""
    This script is executed by the `software-packages` command.

    The purpose of this script is to play the integration role between
    software-packages and PatchRunner. It is responsible to call PatchRunner
    and transform its output to the input expected by software-packages.
"""
import sys
from os.path import join, exists, abspath, dirname, isdir
import os
import logging
import subprocess
from time import sleep
from datetime import datetime
import tempfile
import time
import shutil

from vmware.update.json_utils import prettyJsonDump, JsonFileSerializer, JsonSerializer
from vmware.update.extensions import extend, Hook
from vmware.update.specs import RequirementsResult, Requirements, \
        ValidationResult, Mismatch, InternalError, StatusInfo, ErrorInfo, \
        PatchInfo, Question
from vmware.update.l10n import msgMetadata as _T, localizedString as _

logger = logging.getLogger(__name__)

DEPLOYMENT_ROOT = abspath(join(dirname(__file__), os.pardir))
DEPLOYMENT_TYPE_FILE = "/etc/vmware/deployment.node.type"
IGNORELISTPATTERN = "%s_ignore.lst"

INTERNAL_ERROR_TEXT = _(_T('ur.internal.text',
                           'Internal error occurs during execution of update process.'))
INTERNAL_ERROR_RESOLUTION = _(_T('ur.internal.resolution',
                                 'Send upgrade log files to VMware technical support '
                                 'team for further assistance.'))
PATCH_SUMMARY = _(_T('patch.summary',
                     'Test patch summary for testing only.'))
TIME_TO_INSTALL= 0

def _getRPMBlacklist():
    """Retrieves the rpm blacklist based on the current deployment type.

    @return: Rpms which should not be installed during the update workflow
    @rtype: list
    """
    deployType = ""
    if exists(DEPLOYMENT_TYPE_FILE):
        with open(DEPLOYMENT_TYPE_FILE) as f:
            deployType = f.read().strip()

    logger.debug("vCSA deployment Type: %s", deployType)
    ignoreFile = join(DEPLOYMENT_ROOT, "ignorelists",
                      IGNORELISTPATTERN % (deployType))

    ignoreRpms = []
    if exists(ignoreFile):
        with open(ignoreFile) as f:
            ignoreRpms = [line.strip() for line in f if line.strip()]
    else:
        logger.debug("File %s does not exist", ignoreFile)

    return ignoreRpms

def _backspaceStdout(l):
    """ Go back l characters in stdout and replace them with spaces
    """

    sys.stdout.write("%s%s%s" % (("\b" * l), (" " * l), ("\b" * l)))
    sys.stdout.flush()

def _getFormattedTime():
    """ Output the current time in a format consistent with
        the logging.
        example: [2014-07-24 08:42:37,084]
    """

    d = datetime.now()
    return ("%04d-%02d-%02d %02d:%02d:%02d,%03d" % \
        (d.year, d.month, d.day, d.hour, d.minute, d.second, d.microsecond/1000))

def _runCommand(command, progress = False, message = "", onSubcmdStart=None):
    """Run a command and waits until its completion.

    @param command: The command to run
    @type command: List

    @param progress: Flag indicating whether there should be a progress bar
    @type progress: boolean

    @param message: The message to be displayed if the ahead of the progress bar
    @type message: str

    @param onSubcmdStart: Executed when a subcommand is started
    @type onSubcmdStart: Callback
      Example: def onSubcmdStart():

    @return: Tuple of (stdout, stderr, processExitCode)
    @rtype: tuple
    """

    logger.debug("Running command: %s", command)
    proc = subprocess.Popen(command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, stdin=subprocess.PIPE)

    if onSubcmdStart:
        onSubcmdStart()

    if progress:
        # This code creates a progress bar and polls to see if the
        # process has completed, once the progress bar has finished.
        # If the process has completed the return with the error code,
        # otherwise repeat the progress bar.
        progressbar_width = 4
        progressbar_time = 1.0
        # setup progress bar
        sys.stdout.write(" [%s] : %s%s" % (_getFormattedTime(), message, \
            (" " * progressbar_width)))
        sys.stdout.flush()
        while (proc.poll() == None):
            logger.debug("STDOUT: %s" % proc.stdout.read())
            logger.debug("STDERR: %s" % proc.stderr.read())
            _backspaceStdout(progressbar_width)
            for _i in range(progressbar_width):
                sleep(progressbar_time/progressbar_width)
                sys.stdout.write(".")
                sys.stdout.flush()
        sys.stdout.write("\n")
        sys.stdout.flush()

    out, err = proc.communicate()
    rc = proc.returncode
    return (out, err, rc)

def _executePatchRunnerAndSupplyUserData(command, stageDir, userData, outputFile,
                                         reportCompletion):
    '''Execute PatchRunner subcommand and supply user data to executed process.
    The routine will store the user data into a file and will pass the file
    to PatchRunner subcommand, then the file will be destroyed as soon as
    the PatchRunner loads it.

    @param command: PatchRunner command. Could be validate, prepatch, patch
    @type command: str

    @param stageDir: PatchRunner stage directory
    @type stageDir: str

    @param userData: Supplied user data as lookup of user data id and answer
    @type userData: dict

    @param outputFile: Output file passed to PatchRunner subcommand. This will
      be the file where the PatchRunner will store its results
    @type outputFile: str

    @param reportCompletion: Whether the routine needs to report the status of
      the command based on the PatchRunner subcommand exit code
    @type reportCompletion: bool
    '''
    patchRunner = join(DEPLOYMENT_ROOT, "scripts", "patches", "PatchRunner.py")

    with tempfile.NamedTemporaryFile() as fp:
        # Dump answers to a file}
        cnt = JsonSerializer().serialize(userData)
        fp.write(cnt)
        # Flush all the data, so the other process can read it.
        fp.flush()

        cmd = [sys.executable, patchRunner, command, "-u", fp.name,
               "-o", outputFile, '-d', stageDir, '--disableStdoutLogging']
        out, err, rc = _runCommand(cmd,
                                   progress=True,
                                   message="Running %s script." % command)

    logger.debug("Patch command %s: \n out=%s\nerror=%s\nrc=%d\n" %
                 (command, out, err, rc))

    if reportCompletion:
        if rc:
            logger.error("Patch command %s failed", command)
            _reportFailure(outputFile)
        else:
            logger.debug("Patch command %s completed successfully", command)
            _reportSuccess(outputFile)

    return rc == 0

def _reportSuccess(outputFile):
    if exists(outputFile):
        serializer = JsonFileSerializer(outputFile)
        data = serializer.deserialize()
        if data["status"] == StatusInfo.State.SUCCESS:
            logger.debug("Success status has been already reported")
        elif data["status"] == StatusInfo.State.ERROR:
            logger.warn("Unable to report success, because failure has been "
                        "already reported.")
        else:
            logger.debug("Report success")
            data["status"], data["progress"] = StatusInfo.State.SUCCESS, 100
            prettyJsonDump(outputFile, data)
    else:
        logger.debug("Report success")
        result = StatusInfo(StatusInfo.State.SUCCESS, progress=100)
        prettyJsonDump(outputFile, result)

def _reportFailure(outputFile):
    error=ErrorInfo(detail=[INTERNAL_ERROR_TEXT], resolution=INTERNAL_ERROR_RESOLUTION)
    if exists(outputFile):
        serializer = JsonFileSerializer(outputFile)
        data = serializer.deserialize()
        if data["status"] == StatusInfo.State.ERROR:
            logger.debug("Error status has been already reported")
            return

        if data["status"] == StatusInfo.State.SUCCESS:
            logger.warn("Wrong success status has been already reported. "
                        "Overwriting it.")

        logger.debug("Report error")
        data["status"] = StatusInfo.State.ERROR
        data["error"] = error
        prettyJsonDump(outputFile, data)
    else:
        logger.debug("Report error")
        result = StatusInfo(StatusInfo.State.ERROR, error=error)
        prettyJsonDump(outputFile, result)

def _reportProgress(outputFile, progress):
    result = StatusInfo(StatusInfo.State.RUNNING, progress=progress)
    prettyJsonDump(outputFile, result)

def _transformRequirements(requirements, defaultRpmIgnoreList):
    '''Transforms requirements from PatchRunner to requirements defined in
    patching spec. Add rpmIgnoreList based on the deployment type, on top of
    rpmIgnoreList defined by the components

    @param requirements: Aggregated component requirements:
    @type requirements: dict representing Requirements object
     Example:
         {
            "rebootRequired": false,
            "requiredDiskSpace": {
                "core": 0.1
            },
            "rpmIgnoreList": [],
            "questions": [
                {
                    "defaultAnswer": null,
                    "kind": "password",
                    "description": {
                        "id": "sso.password.desc",
                        "localized": "SSO root password is ..",
                        "translatable": "SSO root password is .."
                    },
                    "text": {
                        "id": "sso.password.text",
                        "localized": "What is the SSO root password",
                        "translatable": "What is the SSO root password"
                    },
                    "allowedValues": [],
                    "userDataId": "sso.root.password"
                }
            ]
        }
    @param defaultRpmIgnoreList: Default rpm ignore list which needs to be
      added on top of component requirements
    @type defaultRpmIgnoreList: list

    @return: Patch requirements
    @rtype update_specs.Requirements
    '''
    requiredDiskSpace = requirements['requiredDiskSpace']
    rpmIgnoreList = requirements['rpmIgnoreList'] + defaultRpmIgnoreList
    questions = []
    rebootRequired = requirements['rebootRequired']

    for q in requirements['questions']:
        question = Question(userDataId = q['userDataId'],
                            text = q['text'],
                            description = q['description'],
                            kind = q['kind'],
                            allowedValues = q['allowedValues'],
                            defaultAnswer = q['defaultAnswer'])
        questions.append(question)
    return Requirements(requiredDiskSpace=requiredDiskSpace,
                        rpmIgnoreList=rpmIgnoreList,
                        questions=questions,
                        rebootRequired=rebootRequired)

def _transformPatchInfo(patchInfo):
    '''Transforms patch info returned by PatchRunner to patch info defined in
    patching spec. Replace patch summary with patch summary defined on top
    level.

    @param patchInfo: Aggregated patch info.
    @type patchInfo: dict representing PatchInfo object
     Example:
         {
            "patchSummary": null,
            "timeToInstall": 5
         },

    @return: Patch information
    @rtype update_specs.PatchInfo
    '''
    return PatchInfo(timeToInstall=patchInfo['timeToInstall'],
                     patchSummary=PATCH_SUMMARY)

def _transformMismatches(mismatches):
    """Transforms mismatches from PatchRunner to mismatches defined
    in patching spec. Add given mismatches on top of mismatches returned by
    PatchRunner

    @param mismatches: Patch Runner mismatches
    @type mismatches: array of dict
    """
    result = []
    for mismatch in mismatches:
        result.append(Mismatch(
                text = mismatch['text'],
                description = mismatch['description'],
                resolution = mismatch['resolution'],
                problemId = mismatch['problemId'],
                severity = mismatch['severity'],
                relatedQuestionId = mismatch['relatedUserDataId'])
        )

    return result

def _getPatchRunnerStageDir(rootStageDir):
    return join(rootStageDir, "patch_runner")

@extend(Hook.Requirements)
def collectRequirements(requirementsSpec):
    patchRunnerStageDir = _getPatchRunnerStageDir(requirementsSpec.stageDirectory)
    patchRunner = join(DEPLOYMENT_ROOT, "scripts", "patches", "PatchRunner.py")
    patchRunnerComponentsDir = join(DEPLOYMENT_ROOT, "scripts", "patches", "payload", "components-script")

    if not exists(patchRunnerStageDir):
        os.makedirs(patchRunnerStageDir)

    if requirementsSpec.userData.get('thirdPartyPatch'):
        #This is 3rd party update and only system scripts need to be executed
        for dr in os.listdir(patchRunnerComponentsDir):
            fullPath = join(patchRunnerComponentsDir, dr)
            if isdir(fullPath) and 'first_component' not in dr \
                and 'last_component' not in dr:
                logger.debug('Third party patching, thus removing components directory %s', fullPath)
                shutil.rmtree(fullPath)

    with tempfile.NamedTemporaryFile() as fp:
        cmd = [sys.executable, patchRunner, 'discovery', "-o", fp.name,
               "-d", patchRunnerStageDir, '--disableStdoutLogging']
        out, err, rc = _runCommand(cmd,
                                   progress=True,
                                   message="Running requirements script.")
        logger.debug("Patch command %s: \n out=%s\nerror=%s\nrc=%d\n" %
             (cmd, out, err, rc))

        # There is python bugs on the appliance and the code below does not work
        # properly
        #fp.seek(0)
        #requirementsResult = fp.read()
        with open(fp.name) as tempFp:
            requirementsResult = tempFp.read()

    rpmIgnoreList = _getRPMBlacklist()

    if requirementsResult:
        logger.debug("Transform patch runner requirements results: %s",
             requirementsResult)
        patchRunnerRes = JsonSerializer().deserialize(requirementsResult)
        requirements = _transformRequirements(patchRunnerRes['requirements'],
                                              rpmIgnoreList)
        patchInfo = _transformPatchInfo(patchRunnerRes['patchInfo'])
        mismatches = _transformMismatches(patchRunnerRes['mismatches'])
        return RequirementsResult(requirements=requirements,
                                  patchInfo=patchInfo,
                                  mismatches=mismatches)
    else:
        if rc:
            logger.error("Requirements failed. Generate error result")
            mismatch = Mismatch(INTERNAL_ERROR_TEXT,
                    description=INTERNAL_ERROR_TEXT,
                    resolution=INTERNAL_ERROR_RESOLUTION,
                    severity=Mismatch.ERROR)
            mismatches = [mismatch]
        else:
            logger.debug("Requirements succeeded. Generate success result")
            mismatches = []
        requirements = Requirements(rpmIgnoreList=rpmIgnoreList)
        patchInfo = PatchInfo(TIME_TO_INSTALL, PATCH_SUMMARY)
        return RequirementsResult(requirements=requirements, patchInfo=patchInfo,
                                  mismatches=mismatches)

@extend(Hook.Validation)
def validate(validationSpec):
    patchRunnerStageDir = _getPatchRunnerStageDir(validationSpec.stageDirectory)
    userData = validationSpec.userData or {}
    with tempfile.NamedTemporaryFile() as fp:
        succ = _executePatchRunnerAndSupplyUserData("validate",
                                                    patchRunnerStageDir,
                                                    userData,
                                                    fp.name,
                                                    reportCompletion=False)

        # There is python bugs on the appliance and the code below does not work
        # properly
        #fp.seek(0)
        #validationResult = fp.read()
        with open(fp.name) as tempFp:
            validationResult = tempFp.read()

    if validationResult:
        logger.debug("Transform patch runner validation results: %s",
                     validationResult)
        patchRunnerRes = JsonSerializer().deserialize(validationResult)
        mismatchResult = _transformMismatches(patchRunnerRes['mismatches'])
        result = ValidationResult(mismatches=mismatchResult)
    elif not succ:
        logger.debug("Validate logic failed. Generate default error")
        mismatch = Mismatch(INTERNAL_ERROR_TEXT,
                            description=INTERNAL_ERROR_TEXT,
                            resolution=INTERNAL_ERROR_RESOLUTION,
                            severity=Mismatch.ERROR)
        result = ValidationResult([mismatch])
    else:
        logger.debug("Validate succeeded. Generate success result")
        result = ValidationResult()

    return result

@extend(Hook.Prepare)
def prepare(prepareSpec):
    # This hook is designed to be used in exceptional cases when system rpms need
    # to be installed before the rest of the rpms
    # if you change the system in this hook you need to WARN the user for that
    # as on failure the customer will need to revert via snapshot
    pass

@extend(Hook.Prepatch)
def prepatch(prepatchSpec):
    patchRunnerStageDir = _getPatchRunnerStageDir(prepatchSpec.stageDirectory)
    userData = prepatchSpec.userData or {}
    _executePatchRunnerAndSupplyUserData("prepatch",
                                         patchRunnerStageDir,
                                         userData,
                                         prepatchSpec.outputFile,
                                         reportCompletion=True)

@extend(Hook.Patch)
def patch(patchSpec):
    patchRunnerStageDir = _getPatchRunnerStageDir(patchSpec.stageDirectory)
    userData = patchSpec.userData or {}
    _executePatchRunnerAndSupplyUserData("patch",
                                         patchRunnerStageDir,
                                         userData,
                                         patchSpec.outputFile,
                                         reportCompletion=True)
