#!/usr/bin/env python
# Copyright 2015 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""
    This script is executed by the `software-packages` command.

    The purpose of this script is to modify the content that gets
    staged on the appliance. For most appliances, all the patch
    payload rpms are applicable, so the pre-stage script should be
    very simple.

    All the patching installation related logic should be placed in
    the pre-install.sh and post-install.sh scripts.
"""
import sys
from os.path import join, exists, abspath, dirname
import os
import logging
import subprocess
import time
from datetime import datetime
import json
import spwd
import versionFilter
from vmware.update.json_utils import prettyJsonDump
from vmware.update.extensions import extend, Hook
from vmware.update.specs import RequirementsResult, Requirements, InternalError, \
        StatusInfo, ValidationResult, Mismatch

from vmware.update.l10n import msgMetadata as _T, localizedString as _

logger = logging.getLogger(__name__)

DEPLOYMENT_ROOT = abspath(join(dirname(__file__), os.pardir))
DEPLOYMENT_TYPE_FILE = "/etc/vmware/deployment.node.type"
IGNORELISTPATTERN = "%s_ignore.lst"
curfilePath = abspath(join(dirname(__file__), os.pardir))
PATCH_HEADER_CONF = os.path.join(curfilePath, 'headers', 'va_patchheader.json')
sys.path.append('/usr/lib/applmgmt/base/py/')
from vmware.vherd.base import update_utils


def _checkVersion():
    ''' Check if base version is greater than
        the patch.
    '''

    with open(update_utils.UPDATE_CONF, "r") as updateConf:
        updateConfJson = json.load(updateConf)
    baseVersion= updateConfJson['version']

    with open(PATCH_HEADER_CONF, "r") as pJson:
        patchHeaderJson = json.load(pJson)
    patchVersion = patchHeaderJson['version']

    patchVersionList = patchVersion.split(".")[:]
    buildVersionList = baseVersion.split(".")[:]
    for i in range(len(patchVersionList)):
        if int(patchVersionList[i]) < int(buildVersionList[i]):
            return False
    rc = versionFilter.filter()
    if rc == 1:
        raise InternalError("Unsupported patch")
    return True

def _getRPMBlacklist():
    """Retrieves the rpm blacklist based on the current deployment type.

    @return: Rpms which should not be installed during the update workflow
    @rtype: list
    """
    deployType = ""
    if exists(DEPLOYMENT_TYPE_FILE):
        with open(DEPLOYMENT_TYPE_FILE) as f:
            deployType = f.read().strip()

    logger.debug("vCSA deployment Type: %s", deployType)
    ignoreFile = join(DEPLOYMENT_ROOT, "ignorelists",
                      IGNORELISTPATTERN % (deployType))

    ignoreRpms = []
    if exists(ignoreFile):
        with open(ignoreFile) as f:
            ignoreRpms = [line.strip() for line in f if line.strip()]
    else:
        logger.debug("File %s does not exist", ignoreFile)

    return ignoreRpms


def _backspaceStdout(l):
    """ Go back l characters in stdout and replace them with spaces
    """

    sys.stdout.write("%s%s%s" % (("\b" * l), (" " * l), ("\b" * l)))
    sys.stdout.flush()


def _getFormattedTime():
    """ Output the current time in a format consistent with
        the logging.
        example: [2014-07-24 08:42:37,084]
    """

    d = datetime.now()
    return ("%04d-%02d-%02d %02d:%02d:%02d,%03d" %
            (d.year, d.month, d.day, d.hour, d.minute, d.second,
             d.microsecond / 1000))


def _runCommand(command, progress=False, message=''):
    """Run a command and waits until its completion.

    @param command: The command to run
    @type command: List

    @param progress: Flag indicating whether there should be a progress bar
    @type progress: boolean

    @param message: The message to be displayed if the ahead of the progress bar
    @type message: str

    @return: Tuple of (stdout, stderr, processExitCode)
    @rtype: tuple
    """

    logger.debug("Running command: %s", command)
    proc = subprocess.Popen(command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, stdin=subprocess.PIPE)

    if progress:
        # This code creates a progress bar and polls to see if the
        # process has completed, once the progress bar has finished.
        # If the process has completed the return with the error code,
        # otherwise repeat the progress bar.
        progressbar_width = 4
        progressbar_time = 1.0
        # setup progress bar
        sys.stdout.write(" [%s] : %s%s" %
                         (_getFormattedTime(), message,
                          (" " * progressbar_width)))
        sys.stdout.flush()
        while (proc.poll() is None):
            logger.debug("STDOUT: %s" % proc.stdout.read())
            logger.debug("STDERR: %s" % proc.stderr.read())
            _backspaceStdout(progressbar_width)
            for _i in range(progressbar_width):
                time.sleep(progressbar_time / progressbar_width)
                sys.stdout.write(".")
                sys.stdout.flush()
        sys.stdout.write("\n")
        sys.stdout.flush()

    out, err = proc.communicate()
    rc = proc.returncode
    return (out, err, rc)


@extend(Hook.Requirements)
def collectRequirements(_requirementsSpec):
    rpmIgnoreList = _getRPMBlacklist()
    requirements = Requirements(rpmIgnoreList=rpmIgnoreList)
    return RequirementsResult(requirements=requirements)


@extend(Hook.Validation)
def validate(_vlidationSpec):
    invalid_patch_message = _(
        _T('com.vmware.applmgmt.no.install.needed',
           'Latest updates already installed on vCSA,'
           ' nothing to stage/install'))
    mismatches = []
    if _checkVersion():
        # Base version is lower
        logger.info('The version check is passed.')
    else:
        mismatches.append(Mismatch(text=invalid_patch_message))

    password_message = _(
        _T('com.vmware.applmgmt.password.expired',
           'Appliance (OS) root password is expired or is going to expire soon.'
           ' Please change the root password before installing an update.'))
    pwdCfg = spwd.getspnam('root')
    maxdays = pwdCfg[4]
    if maxdays != -1:
        if time.time() > (maxdays + pwdCfg[2]) * 86400:
            mismatches.append(Mismatch(text=password_message))
    if mismatches:
        return ValidationResult(mismatches=mismatches)
    return None

def _reportSuccess(outputFile):
    result = StatusInfo(StatusInfo.State.SUCCESS, progress=100)
    prettyJsonDump(outputFile, result)


def _reportFailure(outputFile):
    result = StatusInfo(StatusInfo.State.ERROR)
    prettyJsonDump(outputFile, result)


def _reportProgress(outputFile, progress):
    result = StatusInfo(StatusInfo.State.RUNNING, progress=progress)
    prettyJsonDump(outputFile, result)


@extend(Hook.Prepare)
def prepare(prepareSpec):
    # This hook is designed to be used in exceptional cases when system rpms
    # need to be installed before the rest of the rpms if you change the system
    # in this hook you need to WARN the user for that as on failure the customer
    # will need to revert via snapshot
    pass


@extend(Hook.Prepatch)
def prepatch(prepatchSpec):
    _reportProgress(prepatchSpec.outputFile, 0)
    prepatchScript = join(DEPLOYMENT_ROOT, "scripts", "pre-install.sh")
    out, err, rc = _runCommand([prepatchScript], progress=True,
                               message="Running pre-install script.")
    logger.debug("Prepatch script \n out=%s\nerror=%s\nrc=%d\n" %
                 (out, err, rc))
    if rc:
        _reportFailure(prepatchSpec.outputFile)
        raise InternalError("Unable to successfully execute pre-install.sh")

    logger.info("All VMware services are stopped.")
    logger.debug("Pre-patching completed successfully")
    _reportSuccess(prepatchSpec.outputFile)


@extend(Hook.Patch)
def patch(patchSpec):
    _reportProgress(patchSpec.outputFile, 0)
    prepatchScript = join(DEPLOYMENT_ROOT, "scripts", "post-install.sh")
    out, err, rc = _runCommand([prepatchScript], progress=True,
                               message="Running post-install script.")
    logger.debug("Patch script \n out=%s\nerror=%s\nrc=%d\n" %
                 (out, err, rc))

    if rc:
        _reportFailure(patchSpec.outputFile)
        raise InternalError("Unable to successfully execute post-install.sh")

    logger.debug("Patching completed successfully")
    _reportSuccess(patchSpec.outputFile)
