from subprocess import Popen, PIPE
from const import (
               XML_CFG_FILE, SYMKEY_FILE, PSQL,
               VPXD_CFG_FILE, PYTHON, OPEN_SSL, RUI_KEY_FILE,
               LIBODBC_SO_2, DBC_UPGRADE_BINARY, DBC_UPGRADE_FOLDER, LD_PATH)


import os_utils
import base64
import os

import sys
import json
import platform
import logging

from tempfile import mkstemp, mktemp
from patch_specs import DiscoveryResult, ValidationResult, Question, Mismatch, Requirements, PatchInfo,\
    RequirementsResult
from l10n import msgMetadata as _T, localizedString as _
from patch_errors import PermanentError
from reporting import getProgressReporter

import logging
logger = logging.getLogger(__name__)

from const import (PATCH_OUT_LOG, PATCH_ERR_LOG)

VCDB_PAYLOAD_DIR = os.path.dirname(__file__)
LOG_OUT = os.path.join(VCDB_PAYLOAD_DIR, PATCH_OUT_LOG)
LOG_ERR = os.path.join(VCDB_PAYLOAD_DIR, PATCH_ERR_LOG)

def get_db_user_linux():
   '''
   Get vc database user name from source vpxd.cfg

   @return: user name
   '''
   pipe1 = Popen([XML_CFG_FILE, "-f", VPXD_CFG_FILE, "get", "/config/registry/DB/key_2"], stdout = PIPE)
   dbUser = pipe1.communicate()[0].strip()
   pipe1.stdout.close()
   return dbUser

def get_db_pass_linux_ruikey():
   '''
   Get vc database user password from vpxd.cfg using rui.key
   This implementation is used for vCSA 5.5 source.

   @return: password
   '''
   pyCmd = "import sys\nimport base64\ndata = sys.stdin.read()\ndata = data[1:]\ndecBase64 = base64.b64decode(data)\ndata_len = len(decBase64)\nfor i in range(data_len):\n\tsys.stdout.write(decBase64[data_len-i-1])"
   pipe1 = Popen([PYTHON, XML_CFG_FILE, "-f", VPXD_CFG_FILE, "get", "/config/registry/DB/key_3"], stdout = PIPE)
   pipe2 = Popen([PYTHON, "-c", pyCmd], stdin = pipe1.stdout, stdout = PIPE)
   pipe1.stdout.close()
   pipe3 = Popen([OPEN_SSL, "rsautl", "-inkey", RUI_KEY_FILE, "-pkcs", "-decrypt"], stdin = pipe2.stdout, stdout = PIPE)
   pipe2.stdout.close()
   dbPass = pipe3.communicate()[0].strip()
   pipe3.stdout.close()
   return dbPass

def get_db_pass_symkey():
   '''
   Get vc database user password from vpxd.cfg using symkey.dat
   This implementation is used for vCSA 5.5 source.

   @return: password
   '''
   dbPass = ""
   key3_cmd = [XML_CFG_FILE, "-f", VPXD_CFG_FILE, "get", "/config/registry/DB/key_3"]

   key3_val = os_utils.getCommandOutput(key3_cmd).strip()[1:]
   dbPass = decrypt_key3(get_symkey(), key3_val)
   return dbPass

def decrypt_key3(symkey_hexstring, encrypted_base64):
   '''
   Decrypt password using symkey and encrypted value.

   @symkey_hexstring: symkey.dat contents
   @encrypted_base64: stripped key_3 value from vpxd.cfg
   @return: decrypted password
   '''
   from Crypto.Cipher import AES

   symkey = symkey_hexstring.decode('hex')
   encrypted_raw = base64.b64decode(encrypted_base64)
   assert (len(encrypted_raw) > 16), "Wrong encrypted string length."
   ciphertext = encrypted_raw[16:]
   iv = encrypted_raw[0:16]
   cipher = AES.new(symkey, AES.MODE_CBC, iv)
   try:
      padded_plaintext = cipher.decrypt(ciphertext)
   except Exception, e:
      logger.error("Failed to decrypt password: %s", e)
      error = _(_T('vcdb.error.decrypt.password', "Failed to decrypt password: %s"), e)
      raise PermanentError(cause=error)
   padding_len = ord(padded_plaintext[-1])
   plaintext = padded_plaintext[:-padding_len]

   return plaintext

def get_symkey():
   '''
   Read encryption hex string from symkey.dat

   @return: encryption key
   '''
   with open(SYMKEY_FILE) as f:
      content = f.readlines()
   content_str = content[0].strip()
   symkey_length = len(content_str)
   if symkey_length != 64:
      logger.error("Unexpected symkey length: %d", symkey_length)
      error = _(_T('vcdb.error.symkey.length', "Unexpected symkey length: %d"), symkey_length)
      raise PermanentError(cause=error)
   return content_str


def patch_db(vcUser,vcPass,sourceVcVersion):
   '''
   Run dbcupgrade tool to patch DB
   '''

   if not os.path.islink(LIBODBC_SO_2):
      cmd = ["ln", "-s", "/usr/lib64/libodbc.so.1", LIBODBC_SO_2]
      ret = os_utils.getCommandExitCode(cmd)
      if ret != 0:
            logger.error("Error during execution: %s", cmd)
            error = _(_T('vcdb.db.error.link', 'Error during link libodbc.so.1.'))
            suggestUserAction= _(_T('vcdb.db.resolution.link', 'Check that libodbc.so.1 exists.'))
            raise PermanentError(cause=error, resolution=suggestUserAction)

   cmdPatch = [DBC_UPGRADE_BINARY,
               "-u", base64.b64encode(vcUser),
               "-p", base64.b64encode(vcPass),
               "-b", sourceVcVersion,
               "-s", "0",
               "-y",
               "-t", "postgresql"]

   fileOut = open(LOG_OUT, "w")
   fileError = open(LOG_ERR, "w")
   stdout, stderr, ret = os_utils.executeCommand(cmdPatch,
                                     cwd=DBC_UPGRADE_FOLDER,
                                     env={"LD_LIBRARY_PATH":LD_PATH})
   for s in stdout:
      fileOut.write(s)
   for s in stderr:
      fileError.write(s)

   if ret != 0:
            logger.error("ERROR dbcupgrade")
            error = _(_T('vcdb.db.error.text', 'Error during patch VCDB. See vcdb_patch.err.'))
            suggestUserAction= _(_T('vcdb.db.resolution.text', 'Patch DB is failed. Please check vcdb_patch.log'))
            raise PermanentError(cause=error, resolution=suggestUserAction)

   try:
      cmd = ["unlink", LIBODBC_SO_2]
      os_utils.getCommandExitCode(cmd)
   except Exception:
      logger.info ("Error during unlink %s", LIBOBDC_SO_2)

